## Flex Box Basic

In this tutorial, we will learn how to use ```mui.FlexBox``` to layout components.

We recommend users to read [this tutorial](https://css-tricks.com/snippets/css/a-guide-to-flexbox/) first.

### Flex Box Container
    
```mui.FlexBox``` is a container that can layout components in a flex box. It has the following props:

#### ```flex-direction```

This establishes the **main-axis**, thus defining the direction flex items are placed in the flex container. Flexbox is (aside from optional wrapping) a single-direction layout concept. Think of flex items as primarily laying out either in horizontal rows or vertical columns. Can be ```row | row-reverse | column | column-reverse```. Default is ```row```.

```Python
from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.flex_items = mui.FlexBox([
            mui.Typography("Item 1").prop(border="1px solid red"),
            mui.Typography("Item 2").prop(border="1px solid blue"),
            mui.Typography("Item 3").prop(border="1px solid aqua"),
        ]).prop(flexDirection="row")

        return mui.HBox([
            mui.RadioGroup(["row", "row-reverse", "column", "column-reverse"], self._on_radio_group_changed).prop(row=False),
            self.flex_items.prop(flex=1),
        ]).prop(flex=1)

    async def _on_radio_group_changed(self, value):
        await self.flex_items.send_and_wait(self.flex_items.update_event(flexDirection=value))

```

#### ```flex-wrap```
By default, flex items will all try to fit onto one line. You can change that and allow the items to wrap as needed with this property. Can be ```nowrap | wrap | wrap-reverse```. Default is ```nowrap```.

change ```flex-wrap``` via radio group, then move the resize bar in example to see the effect.

```Python
from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.flex_items = mui.FlexBox([
            mui.Typography("Item 1").prop(border="1px solid red"),
            mui.Typography("Item 2").prop(border="1px solid blue"),
            mui.Typography("Item 3").prop(border="1px solid aqua"),
        ]).prop(flexDirection="row")

        return mui.HBox([
            mui.RadioGroup(["nowrap", "wrap", "wrap-reverse"], self._on_radio_group_changed).prop(row=False),
            mui.Allotment([
                self.flex_items,
                mui.HBox([]),
            ]).prop(defaultSizes=[1, 1])
        ]).prop(width="400px")

    async def _on_radio_group_changed(self, value):
        await self.flex_items.send_and_wait(self.flex_items.update_event(flexWrap=value))

```

#### ```justify-content```

This defines the alignment along the **main axis**. It helps distribute extra free space leftover when either all the flex items on a line are inflexible, or are flexible but have reached their maximum size. It also exerts some control over the alignment of items when they overflow the line. Can be ```flex-start | flex-end | center | space-between | space-around | space-evenly```. Default is ```flex-start```.

```Python

from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.flex_items = mui.FlexBox([
            mui.Typography("Item 1").prop(border="1px solid red"),
            mui.Typography("Item 2").prop(border="1px solid blue"),
            mui.Typography("Item 3").prop(border="1px solid aqua"),
        ]).prop(flexDirection="row")
        options = [
            ("flex-start", "flex-start"),
            ("flex-end", "flex-end"),
            ("center", "center"),
            ("space-between", "space-between"),
            ("space-around", "space-around"),
            ("space-evenly", "space-evenly"),
        ]
        return mui.VBox([
            mui.Select("Justify Content", options, self._on_option_changed),
            self.flex_items.prop(flex=1),
        ]).prop(flex=1)

    async def _on_option_changed(self, value):
        await self.flex_items.send_and_wait(self.flex_items.update_event(justifyContent=value))

```

#### ```align-items```

This defines the default behavior for how flex items are laid out along the **cross axis** on the current line. Think of it as the justify-content version for the cross-axis (perpendicular to the main-axis). Can be ```flex-start | flex-end | center | baseline | stretch```. Default is ```stretch```.

```Python
from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.flex_items = mui.FlexBox([
            mui.Typography("Item 111").prop(border="1px solid red"),
            mui.Typography("Item 22222222").prop(border="1px solid blue"),
            mui.Typography("Item 333333").prop(border="1px solid aqua"),
        ]).prop(flexDirection="column")
        options = [
            ("flex-start", "flex-start"),
            ("flex-end", "flex-end"),
            ("center", "center"),
            ("baseline", "baseline"),
            ("stretch", "stretch"),
        ]
        return mui.VBox([
            mui.Select("Align Items", options, self._on_option_changed),
            self.flex_items.prop(flex=1),
        ]).prop(flex=1)

    async def _on_option_changed(self, value):
        await self.flex_items.send_and_wait(self.flex_items.update_event(alignItems=value))

```

#### ```align-content```

This aligns a flex container’s lines within when there is extra space in the **cross-axis**, similar to how justify-content aligns individual items within the **main-axis**. Can be ```flex-start | flex-end | center | space-between | space-around | stretch```. Default is ```stretch```.

**Note**: This property only takes effect on multi-line flexible containers, where flex-wrap is set to either wrap or wrap-reverse. A single-line flexible container (i.e. where flex-wrap is set to its default value, no-wrap) will not reflect align-content.


```Python
from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.flex_items = mui.FlexBox([
            mui.Typography(f"Item {i}").prop(border="1px solid red") for i in range(20)
        ]).prop(flexDirection="row", flexWrap="wrap")
        options = [
            ("flex-start", "flex-start"),
            ("flex-end", "flex-end"),
            ("center", "center"),
            ("space-between", "space-between"),
            ("space-around", "space-around"),
            ("stretch", "stretch"),
        ]
        return mui.VBox([
            mui.Select("Align Items", options, self._on_option_changed),
            self.flex_items.prop(flex=1),
        ]).prop(flex=1)

    async def _on_option_changed(self, value):
        await self.flex_items.send_and_wait(self.flex_items.update_event(alignContent=value))

```

### Flex Item

Flex item props control the flex behavior of the component in flex parent container.

#### ```flex-grow```

This defines the ability for a flex item to grow if necessary. It accepts a unitless value that serves as a proportion. It dictates what amount of the available space inside the flex container the item should take up. Default is ```0```.

If all items have flex-grow set to 1, the remaining space in the container will be distributed equally to all children. If one of the children has a value of 2, that child would take up twice as much of the space as either one of the others (or it will try, at least).

Negative numbers are invalid.

```Python
from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.item_grow_change = mui.Typography("Item 2").prop(border="1px solid blue", flexGrow=1)
        self.flex_items = mui.FlexBox([
            mui.Typography("Item 1").prop(border="1px solid red", flexGrow=1),
            self.item_grow_change,
            mui.Typography("Item 3").prop(border="1px solid aqua", flexGrow=1),
        ]).prop(flexDirection="column")
        return mui.VBox([
            mui.Slider(0.5, 2.0, 0.05, self._on_option_changed, init_value=1.0),
            self.flex_items.prop(flex=1),
        ]).prop(flex=1)

    async def _on_option_changed(self, value):
        await self.item_grow_change.send_and_wait(self.item_grow_change.update_event(flexGrow=value))

```

#### ```flex-shrink```
This defines the ability for a flex item to shrink if necessary. Default is ```1```.


#### ```flex-basis```

This defines the default size of an element before the remaining space is distributed. It can be a length (e.g. 20%, 5rem, etc.) or a keyword. The auto keyword means “look at my width or height property” (which was temporarily done by the main-size keyword until deprecated). The content keyword means “size it based on the item’s content” – this keyword isn’t well supported yet, so it’s hard to test and harder to know what its brethren max-content, min-content, and fit-content do. Default is ```auto```.

If set to 0, the extra space around content isn’t factored in. If set to auto, the extra space is distributed based on its flex-grow value. See this [graphic](http://www.w3.org/TR/css3-flexbox/images/rel-vs-abs-flex.svg).


#### ```flex```

This is the shorthand for flex-grow, flex-shrink and flex-basis combined. The second and third parameters (flex-shrink and flex-basis) are optional. The default is 0 1 auto, but if you set it with a single number value, like flex: 5;, that changes the flex-basis to 0%, so it’s like setting flex-grow: 5; flex-shrink: 1; flex-basis: 0%;. Default is ```0 1 auto```.

```Python
from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.item_grow_change = mui.Typography("Item 2").prop(border="1px solid blue", flex=1)
        self.flex_items = mui.FlexBox([
            mui.Typography("Item 1").prop(border="1px solid red", flex=1),
            self.item_grow_change,
            mui.Typography("Item 3").prop(border="1px solid aqua", flex=1),
        ]).prop(flexDirection="column")
        return mui.VBox([
            mui.Slider(0.5, 2.0, 0.05, self._on_option_changed, init_value=1.0),
            self.flex_items.prop(flex=1),
        ]).prop(flex=1)

    async def _on_option_changed(self, value):
        await self.item_grow_change.send_and_wait(self.item_grow_change.update_event(flex=value))

```

#### ```align-self```

This allows the default alignment (or the one specified by align-items) to be overridden for individual flex items. Can be ```auto | flex-start | flex-end | center | baseline | stretch```. Default is ```auto```.

Please see the align-items explanation to understand the available values.

Note that float, clear and vertical-align have no effect on a flex item.

```Python

from tensorpc.flow import mui, three, plus, appctx, mark_create_layout

class App:
    @mark_create_layout
    def my_layout(self):
        self.item = mui.Typography("Item 22222222").prop(border="1px solid blue", alignSelf="flex-end")
        self.flex_items = mui.FlexBox([
            mui.Typography("Item 111").prop(border="1px solid red"),
            self.item,
            mui.Typography("Item 333333").prop(border="1px solid aqua"),
        ]).prop(flexDirection="column", alignItems="flex-start")
        options = [
            ("flex-start", "flex-start"),
            ("flex-end", "flex-end"),
            ("center", "center"),
            ("baseline", "baseline"),
            ("stretch", "stretch"),
        ]
        align_self_options = [
            ("auto", "auto"),
            *options
        ]
        return mui.VBox([
            mui.Select("Align Items", options, self._on_option_changed, init_value="flex-start").prop(padding="5px 5px"),
            mui.Select("Align Item-2 Self", align_self_options, self._on_slign_self_option_changed, init_value="flex-end").prop(padding="5px 5px"),
            self.flex_items.prop(flex=1),
        ]).prop(flex=1)

    async def _on_option_changed(self, value):
        await self.flex_items.send_and_wait(self.flex_items.update_event(alignItems=value))
    
    async def _on_slign_self_option_changed(self, value):
        await self.flex_items.send_and_wait(self.item.update_event(alignSelf=value))

```
